package org.apache.lucene.index;

import java.io.IOException;
import java.util.Collection;
import java.util.Map;
import org.apache.lucene.store.IndexOutput;

final class NormsWriter extends InvertedDocEndConsumer {

    @Override
    public void abort() {
    }

    // We only write the _X.nrm file at flush
    void files(Collection<String> files) {
    }

    /** Produce _X.nrm if any document had a field with norms
   *  not disabled */
    @Override
    public void flush(Map<InvertedDocEndConsumerPerThread, Collection<InvertedDocEndConsumerPerField>> threadsAndFields, SegmentWriteState state) throws IOException {
        final Map<FieldInfo, List<NormsWriterPerField>> byField = new HashMap<FieldInfo, List<NormsWriterPerField>>();
        if (!state.fieldInfos.hasNorms()) {
            return;
        }
        // same FieldInfo
        for (final Map.Entry<InvertedDocEndConsumerPerThread, Collection<InvertedDocEndConsumerPerField>> entry : threadsAndFields.entrySet()) {
            final Collection<InvertedDocEndConsumerPerField> fields = entry.getValue();
            final Iterator<InvertedDocEndConsumerPerField> fieldsIt = fields.iterator();
            while (fieldsIt.hasNext()) {
                final NormsWriterPerField perField = (NormsWriterPerField) fieldsIt.next();
                if (perField.upto > 0) {
                    // It has some norms
                    List<NormsWriterPerField> l = byField.get(perField.fieldInfo);
                    if (l == null) {
                        l = new ArrayList<NormsWriterPerField>();
                        byField.put(perField.fieldInfo, l);
                    }
                    l.add(perField);
                } else
                    // Remove this field since we haven't seen it
                    // since the previous flush
                    fieldsIt.remove();
            }
        }
        final String normsFileName = IndexFileNames.segmentFileName(state.segmentName, "", IndexFileNames.NORMS_EXTENSION);
        IndexOutput normsOut = state.directory.createOutput(normsFileName);
        try {
            normsOut.writeBytes(SegmentMerger.NORMS_HEADER, 0, SegmentMerger.NORMS_HEADER.length);
            int normCount = 0;
            for (FieldInfo fi : state.fieldInfos) {
                List<NormsWriterPerField> toMerge = byField.get(fi);
                int upto = 0;
                if (toMerge != null) {
                    final int numFields = toMerge.size();
                    normCount++;
                    final NormsWriterPerField[] fields = new NormsWriterPerField[numFields];
                    int[] uptos = new int[numFields];
                    for (int j = 0; j < numFields; j++) fields[j] = toMerge.get(j);
                    int numLeft = numFields;
                    while (numLeft > 0) {
                        assert uptos[0] < fields[0].docIDs.length : " uptos[0]=" + uptos[0] + " len=" + (fields[0].docIDs.length);
                        int minLoc = 0;
                        int minDocID = fields[0].docIDs[uptos[0]];
                        for (int j = 1; j < numLeft; j++) {
                            final int docID = fields[j].docIDs[uptos[j]];
                            if (docID < minDocID) {
                                minDocID = docID;
                                minLoc = j;
                            }
                        }
                        assert minDocID < state.numDocs;
                        // Fill hole
                        for (; upto < minDocID; upto++) normsOut.writeByte((byte) 0);
                        normsOut.writeByte(fields[minLoc].norms[uptos[minLoc]]);
                        (uptos[minLoc])++;
                        upto++;
                        if (uptos[minLoc] == fields[minLoc].upto) {
                            fields[minLoc].reset();
                            if (minLoc != numLeft - 1) {
                                fields[minLoc] = fields[numLeft - 1];
                                uptos[minLoc] = uptos[numLeft - 1];
                            }
                            numLeft--;
                        }
                    }
                    // Fill final hole with defaultNorm
                    for (; upto < state.numDocs; upto++) normsOut.writeByte((byte) 0);
                } else if (fi.isIndexed && !fi.omitNorms) {
                    normCount++;
                    // Fill entire field with default norm:
                    for (; upto < state.numDocs; upto++) normsOut.writeByte((byte) 0);
                }
                assert 4 + normCount * state.numDocs == normsOut.getFilePointer() : ".nrm file size mismatch: expected=" + (4 + normCount * state.numDocs) + " actual=" + normsOut.getFilePointer();
            }
        } finally {
            normsOut.close();
        }
    }

    /** Produce _X.nrm if any document had a field with norms
   *  not disabled */
    @Override
    public void flush(Map<FieldInfo, InvertedDocEndConsumerPerField> fieldsToFlush, SegmentWriteState state) throws IOException {
        if (!fieldInfos.hasNorms()) {
            return;
        }
        final String normsFileName = IndexFileNames.segmentFileName(state.segmentName, "", IndexFileNames.NORMS_EXTENSION);
        IndexOutput normsOut = state.directory.createOutput(normsFileName);
        try {
            normsOut.writeBytes(SegmentMerger.NORMS_HEADER, 0, SegmentMerger.NORMS_HEADER.length);
            final int numField = fieldInfos.size();
            int normCount = 0;
            for (int fieldNumber = 0; fieldNumber < numField; fieldNumber++) {
                final FieldInfo fieldInfo = fieldInfos.fieldInfo(fieldNumber);
                NormsWriterPerField toWrite = (NormsWriterPerField) fieldsToFlush.get(fieldInfo);
                int upto = 0;
                if (toWrite != null && toWrite.upto > 0) {
                    normCount++;
                    int docID = 0;
                    for (; docID < state.numDocs; docID++) {
                        if (upto < toWrite.upto && toWrite.docIDs[upto] == docID) {
                            normsOut.writeByte(toWrite.norms[upto]);
                            upto++;
                        } else {
                            normsOut.writeByte((byte) 0);
                        }
                    }
                    // we should have consumed every norm
                    assert upto == toWrite.upto;
                    toWrite.reset();
                } else if (fieldInfo.isIndexed && !fieldInfo.omitNorms) {
                    normCount++;
                    // Fill entire field with default norm:
                    for (; upto < state.numDocs; upto++) normsOut.writeByte((byte) 0);
                }
                assert 4 + normCount * state.numDocs == normsOut.getFilePointer() : ".nrm file size mismatch: expected=" + (4 + normCount * state.numDocs) + " actual=" + normsOut.getFilePointer();
            }
        } finally {
            normsOut.close();
        }
    }

    @Override
    void finishDocument() throws IOException {
    }

    @Override
    void startDocument() throws IOException {
    }

    @Override
    InvertedDocEndConsumerPerField addField(DocInverterPerField docInverterPerField, FieldInfo fieldInfo) {
        return new NormsWriterPerField(docInverterPerField, fieldInfo);
    }
}
