package org.apache.lucene.index;

import java.io.IOException;
import org.apache.lucene.document.Fieldable;
import org.apache.lucene.util.ArrayUtil;
import org.apache.lucene.util.RamUsageEstimator;

/** This is a DocFieldConsumer that writes stored fields. */
final class StoredFieldsWriter {

    FieldsWriter fieldsWriter;

    final DocumentsWriterPerThread docWriter;

    int lastDocID;

    int freeCount;

    public StoredFieldsWriter(DocumentsWriter docWriter) {
        this.docWriter = docWriter;
    }

    final DocumentsWriterPerThread.DocState docState;

    public StoredFieldsWriter(DocumentsWriterPerThread docWriter, FieldInfos fieldInfos) {
        this.docWriter = docWriter;
        this.fieldInfos = fieldInfos;
        this.docState = docWriter.docState;
    }

    private int numStoredFields;

    private Fieldable[] storedFields;

    private int[] fieldNumbers;

    public void reset() {
        numStoredFields = 0;
        storedFields = new Fieldable[1];
        fieldNumbers = new int[1];
    }

    public void startDocument() {
        reset();
    }

    public void flush(SegmentWriteState state) throws IOException {
        if (state.numDocs > 0) {
            // It's possible that all documents seen in this segment
            // hit non-aborting exceptions, in which case we will
            // not have yet init'd the FieldsWriter:
            initFieldsWriter();
            fill(state.numDocs);
        }
        if (fieldsWriter != null) {
            fieldsWriter.close();
            fieldsWriter = null;
            lastDocID = 0;
            String fieldsIdxName = IndexFileNames.segmentFileName(state.segmentName, "", IndexFileNames.FIELDS_INDEX_EXTENSION);
            if (4 + ((long) state.numDocs) * 8 != state.directory.fileLength(fieldsIdxName)) {
                throw new RuntimeException("after flush: fdx size mismatch: " + state.numDocs + " docs vs " + state.directory.fileLength(fieldsIdxName) + " length in bytes of " + fieldsIdxName + " file exists?=" + state.directory.fileExists(fieldsIdxName));
            }
        }
    }

    private synchronized void initFieldsWriter() throws IOException {
        if (fieldsWriter == null) {
            fieldsWriter = new FieldsWriter(docWriter.directory, docWriter.getSegment());
            lastDocID = 0;
        }
    }

    int allocCount;

    void abort() {
        reset();
        if (fieldsWriter != null) {
            fieldsWriter.abort();
            fieldsWriter = null;
            lastDocID = 0;
        }
    }

    /** Fills in any hole in the docIDs */
    void fill(int docID) throws IOException {
        // that had no stored fields:
        while (lastDocID < docID) {
            fieldsWriter.skipDocument();
            lastDocID++;
        }
    }

    void finishDocument() throws IOException {
        assert docWriter.writer.testPoint("StoredFieldsWriter.finishDocument start");
        initFieldsWriter();
        fill(docState.docID);
        if (fieldsWriter != null && numStoredFields > 0) {
            fieldsWriter.startDocument(numStoredFields);
            for (int i = 0; i < numStoredFields; i++) {
                fieldsWriter.writeField(fieldNumbers[i], storedFields[i]);
            }
            lastDocID++;
        }
        reset();
        assert docWriter.writer.testPoint("StoredFieldsWriter.finishDocument end");
    }

    public void addField(Fieldable field, FieldInfo fieldInfo) throws IOException {
        if (numStoredFields == storedFields.length) {
            int newSize = ArrayUtil.oversize(numStoredFields + 1, RamUsageEstimator.NUM_BYTES_OBJECT_REF);
            Fieldable[] newArray = new Fieldable[newSize];
            System.arraycopy(storedFields, 0, newArray, 0, numStoredFields);
            storedFields = newArray;
        }
        if (numStoredFields == fieldNumbers.length) {
            fieldNumbers = ArrayUtil.grow(fieldNumbers);
        }
        storedFields[numStoredFields] = field;
        fieldNumbers[numStoredFields] = fieldInfo.number;
        numStoredFields++;
        assert docState.testPoint("StoredFieldsWriterPerThread.processFields.writeField");
    }
}
